#include "ThreadGlobals.hpp"
#include "WatchdCls.hpp"
#include "Settings.hpp"
#include "Types.hpp"
#include "Satistics.hpp"

#include "Assertion.hpp"
#include "BitVector.hpp"
#include "Hash.hpp"
#include "Heap.hpp"
#include "HistoryBuffer.hpp"
#include "IFStream.hpp"
#include "LazyFraction.hpp"
#include "Logging.hpp"
#include "MemPoolA.hpp"
#include "MemPoolP.hpp"
#include "Ptr32.hpp"
#include "OFStream.hpp"
#include "Queue.hpp"
#include "Random.hpp"
#include "VecOfLists.hpp"

#include "AssignInfos.hpp"
#include "BasicSolver.hpp"
#include "ConflictsRestarts.hpp"
#include "Settings.hpp"




const char*  settings_cpp_version(){return "$Revision: 1.9 $";}


Settings::Settings():
    var_decay              	  ( 135 , 1 ), //= 135/128
    var_min_inc				  (1, var_act_t::minNonZeroNumber().getLogOfDenominator() - (2* var_decay.getLogOfDenominator() )),
    var_max_act				  (1000)
{
    Option<bool>::val_type val_true     = std::make_pair<const char*,bool>("yes",true);
    Option<bool>::val_type val_false    = std::make_pair<const char*,bool>("no",false);
    Option<bool>
    b_opt = Option<bool>(&print_console,    false,  "-console=",  "If yes, some information is print to console output");
    b_opt.addPossibleValue(val_true);
    b_opt.addPossibleValue(val_false);
    bool_options.push_back(b_opt);

    b_opt = Option<bool>(&print_values,    true,  "-values=",  			"If yes, a model or an UNSAT core is printed");
	b_opt.addPossibleValue(val_true);
	b_opt.addPossibleValue(val_false);
	bool_options.push_back(b_opt);

    b_opt = Option<bool>(&minimise,    		true,  "-minimise=",  		"If UNSAT, minimise the UNSAT core");
	b_opt.addPossibleValue(val_true);
	b_opt.addPossibleValue(val_false);
	bool_options.push_back(b_opt);

	b_opt = Option<bool>(&bump_agil,    false,  	"-bump-agil=",  		"Bump agil variables");
	b_opt.addPossibleValue(val_true);
	b_opt.addPossibleValue(val_false);
	bool_options.push_back(b_opt);

	b_opt = Option<bool>(&bump_glucose,    true,  	"-bump-glue=",  		"Bump glue variables");
	b_opt.addPossibleValue(val_true);
	b_opt.addPossibleValue(val_false);
	bool_options.push_back(b_opt);


    Option<string>
    s_opt = Option<string>(&csv_file_name,   "",  "-csv-out=",         "If specified, statistics are appended to this file");
    string_options.push_back(s_opt);

    s_opt = Option<string>(&out_file_name,   "",  "-dimacs-out=",      "If specified, core or model is printed to this file");
    string_options.push_back(s_opt);


    Option<uint>
    i_opt = Option<uint>(&rand_seed,      91648253,  "-seed=",           "Specify random seed");
    int_options.push_back(i_opt);

    i_opt = Option<uint>(&timeout,           ~0,     "-timeout=",        "If positive, the solver terminates after the given number of seconds");
    int_options.push_back(i_opt);

    i_opt = Option<uint>(&max_learnts,  20000,  "-max-learnts-init=","Initial number till the first garbage collection");
    int_options.push_back(i_opt);

    i_opt = Option<uint>(&max_learnts_inc,  500,  	"-max-learnts-inc=",	"Increment threshold for garbage collection");
    int_options.push_back(i_opt);

    i_opt = Option<uint>(&pol_mode,  		2,  	"-pol-mode=",			"Polarity Mode: 0=false, 1=true, 2=phase cache");
    int_options.push_back(i_opt);

    i_opt = Option<uint>(&restart_type,  		0,  "-restart-type=",			"Restart type: 0=glue dynamic, 1=luby");
    int_options.push_back(i_opt);

    i_opt = Option<uint>(&lbd_rest_buffr,  		100,  "-restart-buff=",		"Initial value of lbd buffer for restarts");
    int_options.push_back(i_opt);


    i_opt = Option<uint>(&random_var_freq_permil,  20,  "-rand-freq=",		"Frequence of random decisions [permil]");
    int_options.push_back(i_opt);

    i_opt = Option<uint>(&max_lits,  150000000,  	"-max-lits=",			"Maximal number of literals");
    int_options.push_back(i_opt);


}

Settings::~Settings(){}


bool Settings::read(const char* arg){
	const char* value;
	for(uint opt_idx = 0; opt_idx != bool_options.size();++opt_idx){
		if((value = bool_options[opt_idx].canHandle(arg))){
			return bool_options[opt_idx].findValue(value);}
	}

	for(uint opt_idx = 0; opt_idx != int_options.size();++opt_idx){
		if((value = int_options[opt_idx].canHandle(arg))){
			if(!int_options[opt_idx].findValue(value)){
				int_options[opt_idx].assign(atoi(value));}
			return true;
		}
	}

	for(uint opt_idx = 0; opt_idx != string_options.size();++opt_idx){
		if((value = string_options[opt_idx].canHandle(arg))){
			string_options[opt_idx].assign((char*)value);
			return true;}
	}

	if(strncmp(arg, "-version", 8) == 0){
		printCVSVersionNumbers();
		printCurrSettings();
		exit(0);
	}


	if(*arg == '-'){
		cout << "Unknown option: " << arg << endl;
		return false;
	}else if(!cnf_file_name.empty()){
		cout << "One of the two arguments is incorrect: " << cnf_file_name << " or " << arg << endl;
		return false;
	}else {
		cnf_file_name     = string(arg);
	}

	return true;
}

bool Settings::read(int argc, char** argv){

    //check for correct usage
    solver_name = bname(argv[0]);
    if(argc < 2){
        cout << "Not enough Parameters!" << endl;
        return false;
    }

    int index = 0;
    while(++index < argc){
    	if(!read(argv[index])) return false;
    }
    return !cnf_file_name.empty();
};

void Settings::printCurrSettings(){
	cout << std::setfill(' ');

	cout << "Current Values are:" << endl;
    for(uint opt_idx = 0; opt_idx != bool_options.size();++opt_idx){    bool_options[opt_idx].printValue();}
    for(uint opt_idx = 0; opt_idx != string_options.size();++opt_idx){  string_options[opt_idx].printValue();}
    for(uint opt_idx = 0; opt_idx != int_options.size();++opt_idx){     int_options[opt_idx].printValue();}
}


void Settings::printUsage(){

	cout << std::setfill(' ');
	cout << "Program Usage: " << endl;
	cout << solver_name << " [Options] <dimacs-file> " << endl;
	cout << "Options are: " << endl;
	for(uint opt_idx = 0; opt_idx != bool_options.size();++opt_idx){    bool_options[opt_idx].printDescription();}
	for(uint opt_idx = 0; opt_idx != string_options.size();++opt_idx){  string_options[opt_idx].printDescription();}
	for(uint opt_idx = 0; opt_idx != int_options.size();++opt_idx){     int_options[opt_idx].printDescription();}
	cout << endl;

	printCurrSettings();
}



#define HEADER(x,v) std::cout << std::setw(30) << x<<".hpp:   " << v##_hpp_version << endl;
#define SOURCE(x,v) std::cout << std::setw(30) << x<<".cpp:   " << v##_cpp_version() << endl;
#define HEADERSOURCE(x,v) HEADER(x,v) SOURCE(x,v)

void Settings::printCVSVersionNumbers(){

	std::cout << endl;
	std::cout << "MoUsSaka" << endl;
#ifdef NDEBUG
	std::cout << "CVS Version Numbers (RELEASE Compilation)" << endl;
#else
	std::cout << "CVS Version Numbers (DEBUG Compilation)" << endl;
#endif
//global
	SOURCE("Main",main);
	HEADER("Globals",globals);
	HEADER("Satistics",satistics);
	HEADERSOURCE("Settings",settings);
    HEADER("Types",types);

//cdcl
    HEADER("AssignInfos",assigninfos);
    HEADERSOURCE("BasicSolver",basicsolver);
    HEADER("ConflictsRestarts",conflictsrestarts);
    HEADER("Watching",watching);

//simpl

    //HEADERSOURCE("Assym",assym);
    //HEADER("ExtCls",extcls);
    //HEADERSOURCE("SimplyFire",simplyfire);

//util
    HEADER("Assertion",assertion);
	HEADER("BitVector",bitvector);
	HEADER("Hash",hash);
	HEADER("Heap",heap);
	HEADER("HistoryBuffer",historybuffer);
	HEADER("IFStream",ifstream);
	HEADER("LazyFraction",lazyfraction);
	HEADER("Logging",logging);
	HEADER("MemPoolA",mempoola);
	HEADER("MemPoolP",mempoolp);
	HEADER("Ptr32",ptr32);
	HEADER("OFStream",ofstream);
	HEADER("Queue",queue);
	HEADER("Random",random);
	HEADER("VecOfLists",vecoflists);

    //print_out("%s\n",utility_hpp_version);
    printSizes();
}


#define SIZEOF(x) std::cout << std::setw(30) << "Size of " << #x << ": " << sizeof(x) << endl;
void Settings::printSizes(){
	std::cout << endl << "Class Sizes " << endl;
	SIZEOF(void*)

	SIZEOF(ClsP)
	SIZEOF(CemeLit)
	SIZEOF(CoreCand)
	SIZEOF(Reason)
	SIZEOF(WtchEle)
	SIZEOF(TrBool)
	SIZEOF(Variable)
	SIZEOF(Literal)
	SIZEOF(Cls)
	SIZEOF(ClsLit)

	cout << endl;
}






