#ifndef IFSTREAM_HPP_
#define IFSTREAM_HPP_
#define ifstream_hpp_version "$Revision: 1.2 $"


#include <stdio.h>
#include <ctype.h>
#include <stdint.h>
#include <stdlib.h>

/**
 * Class to read from an ASCI-File.
 */ 
class IFStream{
    /** The current character */
    char    cur;
    /** Pointer to the file to read from */
    FILE*   in;
    
    
public:
    /** Initialisation with the name of the file to read.
     * @param file_name: name of the file*/
    inline IFStream(const char* file_name);
    
    /** Close the underlying file.*/
    inline ~IFStream();
    
    /** Checks whether there is a next non-whitespace character. 
     * The cursor is moved to this character. 
     * @return true, if there is a non-whitespace character.*/
    inline bool hasNext();
    
    /** Close the underlying file.*/
    inline void close();
    
    /** Skip the current line. Moves the cursor to the first character of the next line.*/
    inline void skipLine();
    
    /** Move the internal cursor to the next non-whitespace character.*/
    inline void skipWhitespace();

    /** Checks whether the current character equals c.
     * @param c character to be compared
     * @return true if c equals current internal character*/
    inline bool currEquals(const char& c);
    
    /** Retruns the current character without incrementing the internal cursor.
     * @return the current character.*/
    inline char getCurrent();
    
    /** Checks whether the current character is alpha-numeric.
     * @return true, if the current character is alpha-numeric*/
    inline bool currIsAlNum();  
    
    /** Checks if the file is open for reading.
     * @return true, if the file cannot be read. */
    inline bool isBad();
    
    /** Checks if the given character sequence equals the the next sequence of characters in the file.
     * Increments the internal cursor: Either to the first character that differs from the given sequence or in case of 
     * equal character sequences to the next character behind the sequence.
     * @param chrs: a null-terminated character sequence
     * @return true if the given sequence equals the current sequence in the file.*/
    inline bool operator==(const char* chrs);
    
    /** Increments the internal cursor by one.
     * @return a reference to this.*/
    inline IFStream& operator++();
    
    /** Interprets the next sequence of non-whitespace characters as a signed integer.
     * Negative values have to start with a '-' character, positive values may have the prefix '+'.
     * @return the parsed integer.*/
    inline int  getNextInt();
    
private:
    
    /** Prohibited: Copying of IFStream.*/
    IFStream(const IFStream& cp);
    
    /** Prohibited: Assignment of IFStream.*/
    IFStream& operator=(const IFStream& cp);

};


//Implementation

inline      IFStream::~IFStream()                   { if(in != NULL) fclose(in);}

inline void IFStream::close()                       { if(in != NULL) fclose(in),in = NULL; }

inline bool IFStream::currEquals(const char& c)     { return cur == c;}

inline char IFStream::getCurrent()                  { return cur;}

inline bool IFStream::currIsAlNum()                 { return isalnum(cur);}

inline bool IFStream::isBad()                       { return in == NULL;}

inline bool IFStream::hasNext()                     { skipWhitespace(); return cur != EOF;}


inline IFStream::IFStream(const char* file_name): cur(0){
    in = fopen (file_name, "r");
    if(in != NULL) cur = fgetc(in);
}

    
inline void IFStream::skipWhitespace() {
    while (cur == 32 || (cur >= 9 && cur <= 13)){ cur = fgetc(in);}}


inline void IFStream::skipLine() {
    while(cur != EOF && cur != '\0'){
        if (cur == '\n') { cur = fgetc(in); return; }
        cur = fgetc(in);
    } 
}


inline IFStream& IFStream::operator++(){cur = fgetc(in);return *this;}

inline bool IFStream::operator==(const char* chrs){
    while(*chrs != 0 && cur != EOF){
        if(cur != *chrs) return false;
        ++chrs, cur = fgetc(in);
    }
    return cur != EOF;
}


    
inline int32_t IFStream::getNextInt(){
    int32_t sign    ( 1 );
    skipWhitespace();
        
    if      (cur == '-') sign = -1, cur = fgetc(in);
    else if (cur == '+') cur = fgetc(in);
        
    if (cur < '0' || cur > '9') 
        printf("PARSE ERROR at character %c (ASCI: %u)\n", cur,(uint8_t)cur), exit(EXIT_FAILURE);
    
    int32_t val (cur - '0');
    cur = fgetc(in);
    while (cur >= '0' && cur <= '9'){
        val = val*10 + (cur - '0'), cur = fgetc(in);
    }
    return val * sign; 
}

 


#endif /*IFSTREAM_HPP_*/
